# coding=UTF-8
from __future__ import absolute_import, division, print_function
"""Generate serializers/unserializers/length getters for given property_flags"""
import six
import struct
import logging
from coolamqp.framing.compilation.textcode_fields import get_counter, get_from_buffer, get_serializer
from coolamqp.framing.base import AMQPContentPropertyList
from coolamqp.framing.field_table import enframe_table, deframe_table, frame_table_size


logger = logging.getLogger(__name__)


def _compile_particular_content_property_list_class(zpf, fields):
    """
    Compile a particular content property list.

    Particularity stems from
    :param zpf: zero property list, as bytearray
    :param fields: list of all possible fields in this content property
    """
    from coolamqp.framing.compilation.utilities import format_field_name

    if any(field.basic_type == 'bit' for field in fields):
        return u"raise NotImplementedError('I don't support bits in properties yet')"

    # Convert ZPF to a list of [if_exists::bool]
    even = True
    zpf_bits = []
    for q in bytearray(zpf):
        p = bin(q)[2:]
        p = (u'0' * (8 - len(p))) + p

        if not even:
            p = p[:7]

        zpf_bits.extend(map(lambda x: bool(int(x)), p))

    zpf_length = len(zpf)

    # 1 here does not mean that field is present. All bit fields are present, but 0 in a ZPF. Fix this.
    zpf_bits = [zpf_bit or field.type == 'bit' for zpf_bit, field in zip(zpf_bits, fields)]

    mod = [u'''class ParticularContentTypeList(AMQPContentPropertyList):
    """
    For fields:
''']

    for field in fields:
        mod.append(u'    * %s::%s' % (format_field_name(field.name), field.type))
        if field.reserved:
            mod.append(u' (reserved)')
        mod.append(u'\n')

    x = repr(six.binary_type(zpf))
    if not x.startswith('b'):
        x = 'b'+x

    present_fields = [field for field, present in zip(fields, zpf_bits) if present]

    mod.append(u'''
    """
''')

    if len(present_fields) == 0:
        slots = u''
    else:
        slots = (u', '.join((u"u'%s'" % format_field_name(field.name) for field in present_fields)))+u', '

    mod.append(u'''
    __slots__ = (%s)
''' % slots)

    mod.append(u'''
    # A value for property flags that is used, assuming all bit fields are FALSE (0)
    ZERO_PROPERTY_FLAGS = %s
''' % (x, ))

    if len(present_fields) > 0:
        mod.append(u'''
    def __init__(self, %s):
''' % (u', '.join(format_field_name(field.name) for field in present_fields)))

    for field in present_fields:
        mod.append(u'        self.%s = %s\n'.replace(u'%s', format_field_name(field.name)))

    # Let's do write_to
    mod.append(u'\n    def write_to(self, buf):\n')
    mod.append(u'        buf.write(')
    repred_zpf = repr(zpf)
    if not repred_zpf.startswith(u'b'):
        repred_zpf = u'b' + repred_zpf
    mod.append(repred_zpf)
    mod.append(u')\n')

    mod.append(get_serializer(present_fields, prefix=u'self.', indent_level=2))

    # from_buffer
    # note that non-bit values
    mod.append(u'    @classmethod\n')
    mod.append(u'    def from_buffer(cls, buf, start_offset):\n        offset = start_offset + %s\n' % (zpf_length, ))
    mod.append(get_from_buffer(
        present_fields
        , prefix='', indent_level=2))
    mod.append(u'        return cls(%s)\n' %
               u', '.join(format_field_name(field.name) for field in present_fields))


    # get_size
    mod.append(u'\n    def get_size(self):\n')
    mod.append(get_counter(present_fields, prefix=u'self.', indent_level=2)[:-1])    # skip eol
    mod.append(u' + %s\n' % (zpf_length, ))   # account for pf length

    return u''.join(mod)


def compile_particular_content_property_list_class(zpf, fields):
    q = _compile_particular_content_property_list_class(zpf, fields)
    loc = {}
    exec(q, globals(), loc)
    return loc['ParticularContentTypeList']




# coding=UTF-8
from __future__ import absolute_import, division, print_function

from coolamqp.framing.definitions import HARD_ERRORS, SOFT_ERRORS, CONNECTION_FORCED, INVALID_PATH, FRAME_ERROR, \
    SYNTAX_ERROR, COMMAND_INVALID, CHANNEL_ERROR, UNEXPECTED_FRAME, RESOURCE_ERROR, NOT_ALLOWED, NOT_IMPLEMENTED, \
    INTERNAL_ERROR, CONTENT_TOO_LARGE, NO_CONSUMERS, ACCESS_REFUSED, NOT_FOUND, RESOURCE_LOCKED, PRECONDITION_FAILED


class CoolAMQPError(Exception):
    """Base class for CoolAMQP errors"""


class ConnectionDead(CoolAMQPError):
    """
    Operation could be not completed because some other error
    than a legit AMQPError occurred, such as exploding kitten
    """


class AMQPError(CoolAMQPError):
    """
    Base class for errors received from AMQP server
    """

    def is_hard_error(self):
        """Does this error close the connection?"""
        return self.reply_code in HARD_ERRORS

    def __init__(self, *args):
        """

        :param args: can be either reply_code, reply_text, class_id, method_id
                     or a ConnectionClose/ChannelClose.
        """
        from coolamqp.framing.definitions import ConnectionClose, ChannelClose

        if isinstance(args[0], (ConnectionClose, ChannelClose)):
            self.reply_code = args[0].reply_code
            self.reply_text = args[0].reply_text
            self.class_id = args[0].class_id
            self.method_id = args[0].method_id
        else:
            assert len(args) == 4
            self.reply_code, self.reply_text, self.class_id, self.method_id = args



# coding=utf-8
"""
Module used to publish messages.

Expect wild NameErrors if you build this without RabbitMQ extensions (enabled by default),
and try to use MODE_CNPUB.

If you use a broker that doesn't support these, just don't use MODE_CNPUB. CoolAMQP is smart enough
to check with the broker beforehand.
"""
from __future__ import absolute_import, division, print_function

import collections
import logging
import struct
import six
import warnings

from coolamqp.framing.definitions import ChannelOpenOk, BasicPublish, Basic, BasicAck
from coolamqp.framing.frames import AMQPMethodFrame, AMQPBodyFrame, AMQPHeaderFrame

try:
    # these extensions will be available
    from coolamqp.framing.definitions import ConfirmSelect, ConfirmSelectOk, BasicNack
except ImportError:
    pass

from coolamqp.attaches.channeler import Channeler, ST_ONLINE, ST_OFFLINE
from coolamqp.uplink import PUBLISHER_CONFIRMS, MethodWatch, FailWatch
from coolamqp.attaches.utils import AtomicTagger, FutureConfirmableRejectable, Synchronized

from coolamqp.objects import Future, Exchange

logger = logging.getLogger(__name__)


# for holding messages when MODE_CNPUB and link is down
CnpubMessageSendOrder = collections.namedtuple('CnpubMessageSendOrder', ('message', 'exchange_name',
                                                                         'routing_key', 'future'))

#todo what if publisher in MODE_CNPUB fails mid message? they dont seem to be recovered


class Publisher(Channeler, Synchronized):
    """
    An object that is capable of sucking into a Connection and sending messages.
    Depending on it's characteristic, it may process messages in:

        - non-ack mode (default) - messages will be dropped on the floor if there is no active uplink
        - Consumer Publish mode - requires broker support, each message will be ACK/NACKed by the broker
                                  messages will survive broker reconnections.

                                  If you support this, it is your job to ensure that broker supports
                                  publisher_confirms. If it doesn't, this publisher will enter ST_OFFLINE
                                  and emit a warning.

        Other modes may be added in the future.

    Since this may be called by other threads than ListenerThread, this has locking.

    _pub and on_fail are synchronized so that _pub doesn't see a partially destroyed class.
    """
    MODE_NOACK = 0      # no-ack publishing
    MODE_CNPUB = 1      # RabbitMQ publisher confirms extension
    #todo add fallback using plain AMQP transactions - this will remove UnusablePublisher and stuff


    class UnusablePublisher(Exception):
        """This publisher will never work (eg. MODE_CNPUB on a broker not supporting publisher confirms)"""

    def __init__(self, mode):
        """
        Create a new publisher
        :param mode: Publishing mode to use. One of:
                         MODE_NOACK - use non-ack mode
                         MODE_CNPUB - use consumer publishing mode. A switch to MODE_TXPUB will be made
                                      if broker does not support these.
        :raise ValueError: mode invalid
        """
        Channeler.__init__(self)
        Synchronized.__init__(self)

        if mode not in (Publisher.MODE_NOACK, Publisher.MODE_CNPUB):
            raise ValueError(u'Invalid publisher mode')

        self.mode = mode

        self.messages = collections.deque() # Messages to publish. From newest to last.
                                            # tuple of (Message object, exchange name::str, routing_key::str,
                                            #           Future to confirm or None, flags as tuple|empty tuple

        self.tagger = None  # None, or AtomicTagger instance id MODE_CNPUB

        self.critically_failed = False

    @Synchronized.synchronized
    def attach(self, connection):
        Channeler.attach(self, connection)
        connection.watch(FailWatch(self.on_fail))

    @Synchronized.synchronized
    def on_fail(self):
        self.state = ST_OFFLINE

    def _pub(self, message, exchange_name, routing_key):
        """
        Just send the message. Sends BasicDeliver + header + body.

        BECAUSE OF publish THIS CAN GET CALLED BY FOREIGN THREAD.

        :param message: Message instance
        :param exchange_name: exchange to use
        :param routing_key: routing key to use
        :type exchange_name: bytes
        :param routing_key: bytes
        """
        # Break down large bodies
        bodies = []

        body = memoryview(message.body)
        max_body_size = self.connection.frame_max - AMQPBodyFrame.FRAME_SIZE_WITHOUT_PAYLOAD
        while len(body) > 0:
            bodies.append(body[:max_body_size])
            body = body[max_body_size:]


        self.connection.send([
            AMQPMethodFrame(self.channel_id, BasicPublish(exchange_name, routing_key, False, False)),
            AMQPHeaderFrame(self.channel_id, Basic.INDEX, 0, len(message.body), message.properties)
        ])

        # todo optimize it - if there's only one frame it can with previous send
        # no frames will be sent if body.length == 0
        for body in bodies:
            self.connection.send([AMQPBodyFrame(self.channel_id, body)])

    def _mode_cnpub_process_deliveries(self):
        """
        Dispatch all frames that are waiting to be sent

        To be used when  mode is MODE_CNPUB and we just got ST_ONLINE
        """
        assert self.state == ST_ONLINE
        assert self.mode == Publisher.MODE_CNPUB

        while len(self.messages) > 0:
            msg, xchg, rk, fut = self.messages.popleft()

            if fut.cancelled:
                # Ok, don't do this.
                fut.set_cancel()
                continue

            self.tagger.deposit(self.tagger.get_key(), FutureConfirmableRejectable(fut))

            assert isinstance(xchg, (six.binary_type, six.text_type))

            self._pub(msg, xchg, rk)

    def _on_cnpub_delivery(self, payload):
        """
        This gets called on BasicAck and BasicNack, if mode is MODE_CNPUB
        """
        assert self.mode == Publisher.MODE_CNPUB

        if isinstance(payload, BasicAck):
            self.tagger.ack(payload.delivery_tag, payload.multiple)
        elif isinstance(payload, BasicNack):
            self.tagger.nack(payload.delivery_tag, payload.multiple)

    @Synchronized.synchronized
    def publish(self, message, exchange=b'', routing_key=b''):
        """
        Schedule to have a message published.

        If mode is MODE_CNPUB:
            this function will return a Future. Future can end either with success (result will be None),
            or exception (a plain Exception instance). Exception will happen when broker NACKs the message:
            that, according to RabbitMQ, means an internal error in Erlang process.

            Returned Future can be cancelled - this will prevent from sending the message, if it hasn't commenced yet.

        If mode is MODE_NOACK:
            this function returns None. Messages are dropped on the floor if there's no connection.

        :param message: Message object to send
        :param exchange: exchange name to use. Default direct exchange by default. Can also be an Exchange object.
        :type exchange: bytes, str or Exchange instance
        :param routing_key: routing key to use
        :return: a Future instance, or None
        :raise Publisher.UnusablePublisher: this publisher will never work (eg. MODE_CNPUB on Non-RabbitMQ)
        """

        if isinstance(exchange, Exchange):
            exchange = exchange.name.encode('utf8')
        elif isinstance(exchange, six.text_type):
            exchange = exchange.encode('utf8')

        assert isinstance(exchange, six.binary_type)

        # Formulate the request
        if self.mode == Publisher.MODE_NOACK:
            # If we are not connected right now, drop the message on the floor and log it with DEBUG
            if self.state != ST_ONLINE:
                logger.debug(u'Publish request, but not connected - dropping the message')
            else:
                self._pub(message, exchange, routing_key)

        elif self.mode == Publisher.MODE_CNPUB:
            fut = Future()

            #todo can optimize this not to create an object if ST_ONLINE already
            cnpo = CnpubMessageSendOrder(message, exchange, routing_key, fut)
            self.messages.append(cnpo)

            if self.state == ST_ONLINE:
                self._mode_cnpub_process_deliveries()

            return fut
        else:
            raise Exception(u'Invalid mode')

    def on_operational(self, operational):
        state = {True: u'up', False: u'down'}[operational]
        mode = {Publisher.MODE_NOACK: u'noack', Publisher.MODE_CNPUB: u'cnpub'}[self.mode]

        logger.info('Publisher %s is %s', mode, state)

    def on_setup(self, payload):

        # Assert that mode is OK
        if self.mode == Publisher.MODE_CNPUB:
            if PUBLISHER_CONFIRMS not in self.connection.extensions:
                warnings.warn(u'Broker does not support publisher_confirms, refusing to start publisher',
                              RuntimeWarning)
                self.state = ST_OFFLINE
                self.critically_failed = True
                return

        logger.debug('Publisher on_setup, payload=%s', payload)

        if isinstance(payload, ChannelOpenOk):
            # Ok, if this has a mode different from MODE_NOACK, we need to additionally set up
            # the functionality.

            if self.mode == Publisher.MODE_CNPUB:
                self.method_and_watch(ConfirmSelect(False), ConfirmSelectOk, self.on_setup)
            elif self.mode == Publisher.MODE_NOACK:
                # A-OK! Boot it.
                self.state = ST_ONLINE
                self.on_operational(True)

        elif self.mode == Publisher.MODE_CNPUB:
            # Because only in this case it makes sense to check for MODE_CNPUB

            if isinstance(payload, ConfirmSelectOk):
                # A-OK! Boot it.
                self.state = ST_ONLINE
                self.on_operational(True)

                self.tagger = AtomicTagger()

                # now we need to listen for BasicAck and BasicNack

                mw = MethodWatch(self.channel_id, (BasicAck, BasicNack), self._on_cnpub_delivery)
                mw.oneshot = False
                self.connection.watch(mw)
                self._mode_cnpub_process_deliveries()



# coding=UTF-8
"""
Concrete frame definitions
"""
from __future__ import absolute_import, division, print_function

import struct
import six

from coolamqp.framing.base import AMQPFrame
from coolamqp.framing.definitions import FRAME_METHOD, FRAME_HEARTBEAT, FRAME_BODY, FRAME_HEADER, FRAME_END, \
    IDENT_TO_METHOD, CLASS_ID_TO_CONTENT_PROPERTY_LIST, FRAME_METHOD_BYTE, FRAME_BODY_BYTE, FRAME_HEADER_BYTE, \
    FRAME_END_BYTE


class AMQPMethodFrame(AMQPFrame):
    FRAME_TYPE = FRAME_METHOD

    def __init__(self, channel, payload):
        """
        :param channel: channel ID
        :param payload: AMQPMethodPayload instance
        """
        AMQPFrame.__init__(self, channel)
        self.payload = payload

    def write_to(self, buf):
        if self.payload.IS_CONTENT_STATIC:
            buf.write(struct.pack('!BH', FRAME_METHOD, self.channel))
            buf.write(self.payload.STATIC_CONTENT)
        else:
            buf.write(struct.pack('!BHL', FRAME_METHOD, self.channel,
                                  4 + self.payload.get_size()))
            buf.write(self.payload.BINARY_HEADER)
            self.payload.write_arguments(buf)
            buf.write(FRAME_END_BYTE)

    @staticmethod
    def unserialize(channel, payload_as_buffer):
        clsmet = struct.unpack_from('!HH', payload_as_buffer, 0)

        try:
            method_payload_class = IDENT_TO_METHOD[clsmet]
            payload = method_payload_class.from_buffer(payload_as_buffer, 4)
        except KeyError:
            raise ValueError('Invalid class %s method %s' % clsmet)
        else:
            return AMQPMethodFrame(channel, payload)

    def get_size(self):
        # frame_header = (method(1) + channel(2) + length(4) + class(2) + method(2) + payload(N) + frame_end(1))
        return 12 + self.payload.get_size()


class AMQPHeaderFrame(AMQPFrame):
    FRAME_TYPE = FRAME_HEADER

    def __init__(self, channel, class_id, weight, body_size, properties):
        """
        :param channel: channel ID
        :param class_id: class ID
        :param weight: weight (lol wut?)
        :param body_size: size of the body to follow
        :param properties: a suitable AMQPContentPropertyList instance
        """
        AMQPFrame.__init__(self, channel)
        self.class_id = class_id
        self.weight = weight
        self.body_size = body_size
        self.properties = properties

    def write_to(self, buf):
        buf.write(struct.pack('!BHLHHQ', FRAME_HEADER, self.channel,
                              12+self.properties.get_size(), self.class_id, 0, self.body_size))
        self.properties.write_to(buf)
        buf.write(FRAME_END_BYTE)

    @staticmethod
    def unserialize(channel, payload_as_buffer):
        # payload starts with class ID
        class_id, weight, body_size = struct.unpack_from('!HHQ', payload_as_buffer, 0)
        properties = CLASS_ID_TO_CONTENT_PROPERTY_LIST[class_id].from_buffer(payload_as_buffer, 12)
        return AMQPHeaderFrame(channel, class_id, weight, body_size, properties)

    def get_size(self):
        # frame header is always 7, frame end is 1, content header is 12 + props
        return 20 + self.properties.get_size()


class AMQPBodyFrame(AMQPFrame):
    FRAME_TYPE = FRAME_BODY

    FRAME_SIZE_WITHOUT_PAYLOAD = 8

    def __init__(self, channel, data):
        """
        :type data: binary
        """
        AMQPFrame.__init__(self, channel)
        assert isinstance(data, (six.binary_type, memoryview))
        self.data = data

    def write_to(self, buf):
        buf.write(struct.pack('!BHL', FRAME_BODY, self.channel, len(self.data)))
        buf.write(self.data)
        buf.write(FRAME_END_BYTE)

    @staticmethod
    def unserialize(channel, payload_as_buffer):
        return AMQPBodyFrame(channel, payload_as_buffer)

    def get_size(self):
        return 8 + len(self.data)


class AMQPHeartbeatFrame(AMQPFrame):
    FRAME_TYPE = FRAME_HEARTBEAT
    LENGTH = 8
    DATA = struct.pack('!BHLB', FRAME_HEARTBEAT, 0, 0, FRAME_END)

    def __init__(self):
        AMQPFrame.__init__(self, 0)

    def write_to(self, buf):
        buf.write(AMQPHeartbeatFrame.DATA)

    def get_size(self):
        return AMQPHeartbeatFrame.LENGTH



# coding=UTF-8
from __future__ import absolute_import, division, print_function

import math
from collections import namedtuple

import six

from coolamqp.framing.base import BASIC_TYPES, DYNAMIC_BASIC_TYPES

# docs may be None

Constant = namedtuple('Constant', ('name', 'value', 'kind', 'docs'))  # kind is AMQP constant class # value is int
Field = namedtuple('Field', ('name', 'type', 'label', 'docs', 'reserved', 'basic_type'))  # reserved is bool
# synchronous is bool, constant is bool
# repponse is a list of method.name
Class_ = namedtuple('Class_', ('name', 'index', 'docs', 'methods', 'properties'))  # label is int
Domain = namedtuple('Domain', ('name', 'type', 'elementary'))  # elementary is bool


class Method(object):
    def __init__(self, name, synchronous, index, label, docs, fields, response, sent_by_client, sent_by_server):
        self.name = name
        self.synchronous = synchronous
        self.index = index
        self.fields = fields
        self.response = response
        self.label = label
        self.docs = docs
        self.sent_by_client = sent_by_client
        self.sent_by_server = sent_by_server

        self.constant = len([f for f in self.fields if not f.reserved]) == 0

    def get_static_body(self):  # only arguments part
        body = []
        bits = 0
        for field in self.fields:

            if bits > 0 and field.basic_type != 'bit':
                body.append(b'\x00' * math.ceil(bits / 8))
                bits = 0

            if field.basic_type == 'bit':
                bits += 1
            else:
                body.append(eval(BASIC_TYPES[field.basic_type][2]))
        return b''.join(body)

    def is_static(self, domain_to_type=None):  # is size constant?
        for field in self.fields:
            if field.basic_type in DYNAMIC_BASIC_TYPES:
                return False
        return True


def get_size(fields):  # assume all fields have static length
    """Assuming all fields have static length, return their length together. Supports bits"""
    size = 0
    bits = 0
    for field in fields:

        if (bits > 0) and (field.basic_type != 'bit'):  # sync bits
            size += int(math.ceil(bits / 8))
            bits = 0

        if BASIC_TYPES[field.basic_type][0] is None:
            if field.basic_type == 'bit':
                bits += 1
            else:
                size += len(BASIC_TYPES[field.basic_type][2])  # default minimum entry
        else:
            size += BASIC_TYPES[field.basic_type][0]

    if bits > 0:  # sync bits
        size += int(math.ceil(bits / 8))

    return size


def get_docs(elem):
    """Parse an XML element. Return documentation"""
    for kid in elem.getchildren():

        if kid.tag == 'rule':
            return get_docs(kid)

        s = kid.text.strip().split('\n')
        return u'\n'.join([u.strip() for u in s if len(u.strip()) > 0])

    return None


def for_domain(elem):
    """Parse XML document. Return domains"""
    a = elem.attrib
    return Domain(six.text_type(a['name']), a['type'], a['type'] == a['name'])


def for_field(elem):  # for <field> in <method>
    """Parse method. Return fields"""
    a = elem.attrib
    return Field(six.text_type(a['name']), a['domain'] if 'domain' in a else a['type'],
                 a.get('label', None),
                 get_docs(elem),
                 a.get('reserved', '0') == '1',
                 None)


def for_method(elem):  # for <method>
    """Parse class, return methods"""
    a = elem.attrib
    return Method(six.text_type(a['name']), bool(int(a.get('synchronous', '0'))), int(a['index']), a.get('label', None),
                  get_docs(elem),
                  [for_field(fie) for fie in elem.getchildren() if fie.tag == 'field'],
                  [e.attrib['name'] for e in elem.findall('response')],
                  # if chassis=server that means server has to accept it
                  any([e.attrib.get('name', '') == 'server' for e in elem.getchildren() if e.tag == 'chassis']),
                  any([e.attrib.get('name', '') == 'client' for e in elem.getchildren() if e.tag == 'chassis'])
                  )


def for_class(elem):  # for <class>
    """Parse XML, return classes"""
    a = elem.attrib
    methods = sorted([for_method(me) for me in elem.getchildren() if me.tag == 'method'],
                     key=lambda m: (m.name.strip('-')[0], -len(m.response)))
    return Class_(six.text_type(a['name']), int(a['index']), get_docs(elem) or a['label'], methods,
                  [for_field(e) for e in elem.getchildren() if e.tag == 'field'])


def for_constant(elem):  # for <constant>
    """Parse XML, return constants"""
    a = elem.attrib
    return Constant(a['name'], int(a['value']), a.get('class', ''), get_docs(elem))


def get_constants(xml):
    return [for_constant(e) for e in xml.findall('constant')]


def get_classes(xml):
    return [for_class(e) for e in xml.findall('class')]


def get_domains(xml):
    return [for_domain(e) for e in xml.findall('domain')]


def as_unicode(callable):
    def roll(*args, **kwargs):
        return six.text_type(callable(*args, **kwargs))

    return roll


def to_dict_by_name(list_of_things):
    return dict((a.name, a) for a in list_of_things)


@as_unicode
def name_class(classname):
    """Change AMQP class name to Python class name"""
    return classname.capitalize()


@as_unicode
def format_method_class_name(methodname):
    if '-' in methodname:
        i = methodname.find('-')
        return methodname[0:i].capitalize() + methodname[i + 1].upper() + methodname[i + 2:]
    else:
        return methodname.capitalize()


@as_unicode
def format_field_name(field):
    if field in (u'global', u'type'):
        field = field + '_'
    return field.replace('-', '_')


def frepr(p, sop=six.text_type):
    if isinstance(p, (six.binary_type, six.text_type)):
        p = sop(p)
    s = repr(p)

    if isinstance(p, (six.binary_type, six.text_type)) and not s.startswith('u'):
        return ('u' if sop == six.text_type else 'b') + s
    else:
        return s


def to_code_binary(p):
    body = []
    for q in p:
        if isinstance(q, int):
            q = six.int2byte(q)
        z = (hex(ord(q))[2:].upper())
        if len(z) == 1:
            z = u'0' + z
        body.append(u'\\x' + z)
    return u"b'" + (u''.join(body)) + u"'"


def pythonify_name(p):
    return p.strip().replace('-', '_').upper()


def try_to_int(p):
    try:
        return int(p)
    except ValueError:
        return p


def to_docstring(label, doc, prefix=4, blank=True):  # output a full docstring section
    label = [] if label is None else [label]
    doc = [] if doc is None else [q.strip() for q in doc.split(u'\n') if len(q.strip()) > 0]
    pre = u' ' * prefix

    doc = label + doc

    if len(doc) == 0:
        return u''

    doc[0] = doc[0].capitalize()

    if len(doc) == 1:
        return doc[0]

    doc = [p for p in doc if len(p.strip()) > 0]

    if blank:
        doc = [doc[0], u''] + doc[1:]

    f = (u'\n'.join(pre + lin for lin in doc))[prefix:]
    return f


def ffmt(data, *args, **kwargs):
    for arg in args:
        op = str if kwargs.get('sane', True) else frepr
        data = data.replace('%s', op(arg), 1)
    return data
